
## Dependencies

import requests
import pandas as pd
import numpy as np
import time
from datetime import datetime
from bs4  import BeautifulSoup
import re
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests.packages.urllib3.util.retry import Retry

from chickenstats.chicken_nhl.player_names import correct_names_dict, goalie_names, api_names_dict
from chickenstats.chicken_nhl.helper_functions import s_session, convert_ids, convert_to_list, print_scrape_intro
from chickenstats.chicken_nhl.helper_functions import print_concat_finish, print_game_id_time, print_number_of_games, end_timer, hs_strip_html

## Function to scrape schedule
def scrape_schedule(seasons, game_types = ['R', 'P'], final_only = False, live_only = False, _print = True):
    
    """
    Scrapes NHL schedule API and returns a dataframe.
    Seasons should be a single four-digit year or a list of four digit years. Years can be either strings or integers.
    Game types should either be 'all' or a list of desired game types ('PR', 'R', 'P', 'A'). By default will scrape all games. 
    Live games are by default True. If True, includues unplayed games in schedule.
    By default, prints progress to the console.
    """
    
    ## Start time for timer purposes
    scrape_start = time.perf_counter()
    
    ## Starting requests session
    s = s_session()
    
    ## Create list of season IDs. If single season, convert to a list of a single season ID. Else, use list comprehension to convert to season IDs
    seasons = convert_to_list(seasons)
        
    ## Printing start time if printing is enabled
    if _print == True:
        
        number_of_seasons = len(seasons)
        current_time = datetime.now().strftime("%H:%M")
        
        if number_of_seasons > 1:
            number = 'seasons'
            
        else:
            number = 'season'
            
        print(f'Getting schedulue for {number_of_seasons} {number}. Scrape started at {current_time}\n')
    
    
    ## Create blank dataframe to hold all schedules scraped
    season_dfs = []
    
    ## Loop through season IDs and scrape schedule data for each season
    for season in seasons:
        
        ## Create season ID for each season
        season_id = str(int(season)) + str(int(season) + 1)       
        
        ## Start time for timer purposes
        season_start = time.perf_counter()
        
        ## Setting url and calling JSON from API
        url = f'https://statsapi.web.nhl.com/api/v1/schedule?season={season_id}'

        response = s.get(url, timeout = .5).json()
        
        ## Setting up initial season schedule dataframe
        season_df = pd.json_normalize(response['dates'], record_path = 'games', sep = '_')

        season_df['season'] = season_id
        
        ## Removing game types based on function argument. Game types dictionary is function argument. By default, only regular season and playoff game types are included
        if game_types != 'all':
            
            season_df = season_df[np.isin(season_df.gameType, game_types)].copy()
        
        ## Removing unfinished/unplayed games from scheduule information. By default live games are included
        if final_only == True:
            
            season_df = season_df[season_df.status_abstractGameState == 'Final'].copy()
        
        ## Adding individual season schedule to big schedule dataframe
        season_dfs.append(season_df)
        
        ## Printing timer for each season if print is True
        if _print == True:
            
            timer = round(time.perf_counter() - season_start, 2)
            
            print(f'Finished with {season} schedule in {timer} seconds...')
    
    df = pd.concat(season_dfs, ignore_index = True)
    
    ## Changing column names using dictionary and rename function
    new_cols = {'gamePk' : 'game_id',
                'gameType' : 'game_type',
                'link' : 'game_link',
                'gameDate' : 'game_date',
                'status_abstractGameState' : 'game_status',
                'status_detailedState' : 'detailed_game_status',
                'status_startTimeTBD' : 'start_time_TBD',
                'teams_away_leagueRecord_wins' : 'away_team_wins',
                'teams_away_leagueRecord_losses' : 'away_team_losses',
                'teams_away_leagueRecord_ot' : 'away_team_OTL',
                'teams_away_score' : 'away_team_score',
                'teams_away_team_id' : 'away_team_id',
                'teams_away_team_name' : 'away_team_name',
                'teams_away_team_link' : 'away_team_link',
                'teams_home_leagueRecord_wins' : 'home_team_wins',
                'teams_home_leagueRecord_losses' : 'home_team_losses',
                'teams_home_leagueRecord_ot' : 'home_team_OTL',
                'teams_home_score' : 'home_team_score',
                'teams_home_team_id' : 'home_team_id',
                'teams_home_team_name' : 'home_team_name',
                'teams_home_team_link' : 'home_team_link',
                'content_link' : 'game_content_link'}

    df.rename(columns = new_cols, inplace = True)
    
    ## Separating game date column into start time, then making game date a date
    df['start_time'] = pd.to_datetime(df.game_date).dt.time
    df['game_date'] = pd.to_datetime(df.game_date).dt.date
    
    ## Dropping columns that we don't want
    
    columns = ['season', 'game_id', 'game_date', 'game_type', 'game_status', 'home_team_name', 'home_team_id', 'home_team_score',
               'away_team_name', 'away_team_id', 'away_team_score', 'detailed_game_status', 'start_time', 'start_time_TBD',
               'home_team_wins', 'away_team_wins', 'home_team_losses', 'away_team_losses','home_team_OTL', 'away_team_OTL',
               'home_team_link', 'away_team_link', 'game_link', 'game_content_link']
    
    df = df[columns]
    
    
    ## Printing timer for whole scrape if printing is enabled
    if _print == True:
        
        timer = round(time.perf_counter() - scrape_start, 2)
    
        print(f'\nFinished with schedule scrape in {timer} seconds\n')
        
    return df

## Function to scrape HTML rosters
def scrape_html_rosters(game_ids, debug = False, _print = False):
    """
    Game IDs can be a list or for a single game. Returns a dataframe of roster information from the API. By default, prints status update to console.
    This will be nested within a larger roster scraping function but can also be accessed on its own.
    """
    
    scrape_start = time.perf_counter()
    
    game_ids = convert_to_list(game_ids)
    
    if _print == True:
        
        print_scrape_intro(game_ids, 'HTML rosters')
    
    s = s_session()
    
    html_list = []
    
    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        html_season_id, html_game_id = convert_ids(game_id)
    
        html_url = f'http://www.nhl.com/scores/htmlreports/{html_season_id}/RO0{html_game_id}.HTM'
        
        page = s.get(html_url, timeout = .5)
        
        page_status = page.status_code
        
        if page_status == 404:
            
            if debug == True:

                if _print == True:

                    print(f'Sorry, {game_id} does not have any player information')

        else:
            
            soup = BeautifulSoup(page.content.decode('ISO-8859-1'), 'lxml', multi_valued_attributes = None)

            td_dict = {'align':'center', 'class':['teamHeading + border', 'teamHeading + border '], 'width':'50%'}

            teamsoup = soup.find_all('td', td_dict)

            table_dict = {'align':'center', 'border':'0', 'cellpadding':'0', 'cellspacing':'0', 'width':'100%', 'xmlns:ext':''}

            team_names = {}

            team_soup_list = []

            team_list = ['away', 'home']

            html_roster_list = []

            for idx, team in enumerate(team_list):

                team_names.update({team : teamsoup[idx].get_text()})

                team_soup_list.append((soup.find_all('table', table_dict))[idx].find_all('td'))

            for idx, team_soup in enumerate(team_soup_list):

                length = int(len(team_soup) / 3)

                df = pd.DataFrame(np.array(team_soup).reshape(length, 3))

                df.columns = df.iloc[0]

                df = df.assign(team = team_list[idx], team_full_name = team_names.get(team_list[idx]), status = 'active').drop(0)

                html_roster_list.append(df)

            if len(soup.find_all('table', table_dict)) > 2:

                scratch_soups = []

                for idx, team in enumerate(team_list):

                    team_scratch = (soup.find_all('table', table_dict))[idx + 2].find_all('td')

                    if len(team_scratch) > 1:

                        length = int(len(team_scratch) / 3)

                        df = pd.DataFrame(np.array(team_scratch).reshape(length, 3))

                        df.columns = df.iloc[0]

                        df = df.assign(team = team_list[idx], team_full_name = team_names.get(team_list[idx]), status = 'scratch').drop(0)

                        html_roster_list.append(df)

                    else:

                        html_roster_list.append(pd.DataFrame())

            html_game_df = pd.concat(html_roster_list, ignore_index = True)

            html_game_df['game_id'] = game_id
            
            html_game_df['season'] = html_season_id

            html_game_df['team'] = np.where(html_game_df.team == 'CANADIENS MONTREAL', 'MONTREAL CANADIENS', html_game_df.team)

            ## Rename columns
            new_cols = {'#' : 'player_jersey',
                        'Pos' : 'player_position',
                        'status' : 'player_status',
                        'Nom/Name' : 'player_full_name',
                        'Name' : 'player_full_name'}

            html_game_df.rename(columns = new_cols, inplace = True)
            
            if 'player_full_name' not in list(html_game_df.columns):
                
                if debug == True:

                    if _print == True:

                        print(f'Sorry, {game_id} does not have any player information')
            
            else:
                
                ## Fixing position
                positions_dict = {'L' : 'LW', 'R' : 'RW'}

                if 'player_position' not in list(html_game_df.columns):

                    html_game_df['player_position'] = np.nan

                #html_game_df.player_position = html_game_df.player_position.map(positions_dict).fillna(html_game_df.player_position)

                ## Full names, then fixing
                html_game_df.player_full_name = html_game_df.player_full_name.str.split('(').str[0].str.strip()

                # Max Pacioretty doesn't exist in ESPN in 2009-2010, sadly.
                replace_dict = {'AlEXANDRE ' : 'ALEX ',
                                'ALEXANDER ' : 'ALEX ',
                                'CHRISTOPHER ' : 'CHRIS '}

                html_game_df.player_full_name = html_game_df.player_full_name.replace(to_replace = replace_dict)

                # List of names and fixed from Evolving Hockey Scraper.
                html_game_df.player_full_name = html_game_df.player_full_name.map(correct_names_dict).fillna(html_game_df.player_full_name)

                player_names = ['player_first_name', 'player_last_name']

                for idx, player_name in enumerate(player_names):

                    html_game_df[player_name] = html_game_df.player_full_name.str.split(' ', 1).str[idx]

                html_game_df['api_name'] = html_game_df.player_first_name + '.' + html_game_df.player_last_name
                
                
                double_names_fix = {'SEBASTIAN.AHO2' : [html_game_df.api_name == 'SEBASTIAN.AHO', html_game_df.player_position == 'D'],
                                    'COLIN.WHITE2': [html_game_df.api_name == 'COLIN.WHITE', int(html_season_id) >= 20162017], 
                                    'SEAN.COLLINS2' : [html_game_df.api_name == 'SEAN.COLLINS', html_game_df.player_position != 'D'],
                                    'ALEX.PICARD2' : [html_game_df.api_name == 'ALEX.PICARD', html_game_df.player_position != 'D'],
                                    'ERIK.GUSTAFSSON2' : [html_game_df.api_name == 'ERIK.GUSTAFSSON', int(html_season_id) >= 20152016],
                                    'MIKKO.LEHTONEN2' : [html_game_df.api_name == 'MIKKO.LEHTONEN', int(html_season_id) >= 20202021]}

                for fix, conditions in double_names_fix.items():

                    html_game_df.api_name = np.where(np.logical_and(conditions[0], conditions[1]), fix, html_game_df.api_name)

                html_game_df.api_name = np.where(html_game_df.api_name == 'COLIN.', 'COLIN.WHITE2', html_game_df.api_name)

                html_list.append(html_game_df)

            if html_list == []:

                if debug == True:

                    if _print == True:

                        print(f'Sorry, {game_id} does not have any player information...')

            else:

                if _print == True:

                    print_game_id_time(game_start, game_id)
    
    concat_start = time.perf_counter()
    
    if _print == True:
        
        print(f'\nStarting concatenation')
    
    if html_list == []:
        
        html_df = pd.DataFrame()
        
    else:
        
        html_df = pd.concat(html_list, ignore_index = True)
        
        columns = ['season', 'game_id', 'team_full_name', 'team', 'player_full_name',
                   'player_first_name', 'player_last_name', 'api_name', 'player_jersey',
                   'player_position', 'player_status']
        
        columns = [x for x in columns if x in html_df.columns]
        
        html_df = html_df[columns]
        
    if _print == True:

        print_concat_finish(concat_start)

        print_number_of_games(game_ids, scrape_start)
    
    return html_df

## Scrape API events
def scrape_api_events(game_ids, full_scrape = False, debug = False, _print = False):
    """
    Function to scrape play-by-play data from the NHL API. Takes a game ID or a list of game IDs. Does not work for years before 2010.
    
    Can be used on its own, or as part of the larger full scrape function
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    full_scrape -- If False, scrapes the API for each game ID. Give the API response as a JSON otherwise (default False)
    debug -- If True, prints the game ID before attempting to scrape the data (default False)
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)

    if _print == True:
    
        print_scrape_intro(game_ids, 'API events')
    
    ## Important lists
    bad_game_list = list()
    concat_list = list()
    
    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
            
        if full_scrape == False:

            s = s_session()
        
            response = s.get(str('https://statsapi.web.nhl.com/api/v1/game/' + str(game_id) + '/feed/live'), timeout = .5).json()
            
        else:
            
            response = full_scrape
        
        if response['liveData']['plays']['allPlays'] == []:
            
            if debug == True:
            
                print(f"{game_id} doesn't have API info")
                
            bad_game_list.append(game_id)
            
            continue

        ## Creating base pbp dataframe
        event_df = pd.json_normalize(response, sep = '_', record_path = ['liveData', 'plays', 'allPlays'], meta = ['gamePk'])

        ## Getting the event players from the pbp dataframe
        event_players = {x: 'event_player_' + str(x + 1) for x in list(range(0, 4))}

        players = event_df.players.apply(pd.Series).rename(columns = event_players)

        ## Creating a blank list to store the event players for concatenating to the play-level
        players_list = list()

        ## Looping through the event players from the dictionary created earlier
        
        players_loop = [x for x in event_players.values() if x in players.columns]
        
        for player in players_loop:

            ## Creating a dataframe by concatenating the player type with the player's information (unpacking column of dictionaries)
            player_df = pd.concat([players[player].dropna().apply(pd.Series).drop(['player'], axis = 1),
                            players[player].dropna().apply(pd.Series).player.apply(pd.Series)], axis = 1)

            ## Renaming columns
            column_names = {'playerType': player + '_type', 'id': player + '_id',
                            'fullName': player, 'link': player + '_link'}

            player_df.rename(columns = column_names, inplace = True)

            ## Dropping seasonTotal column from the dataframe, but only if it exists
            if 'seasonTotal' in player_df.columns:

                player_df.drop('seasonTotal', axis = 1, inplace = True)

            ## Append the dataframe to the list so it can be concatenated later
            players_list.append(player_df)

        ## Concatenate all of the dataframes in the list on their index
        players_df = pd.concat(players_list, axis = 1)

        ## Concatenate the players dataframe and the pbp dataframe
        event_df = pd.concat([event_df, players_df], axis = 1)

        ## Drop the players column. It's the column of dictionaries that we just unpacked
        event_df = event_df.drop('players', axis = 1)

        ## Change column names        
        column_names = {'result_event': 'event_type_name', 'result_eventCode': 'event_id_nhl', 'result_eventTypeId': 'event_type', 
                        'result_description': 'event_description', 'about_eventIdx': 'event_idx', 'about_eventId': 'event_idx_nhl',
                        'about_period': 'period', 'about_periodType': 'period_type', 'about_ordinalNum': 'period_num',
                        'about_periodTime': 'period_time', 'about_periodTimeRemaining': 'period_time_remaining', 'about_dateTime': 'datetime',
                        'about_goals_away': 'goals_away', 'about_goals_home': 'goals_home', 'coordinates_x': 'coords_x', 'coordinates_y': 'coords_y', 
                        'team_id': 'event_team_id', 'team_name': 'event_team_name', 'team_link': 'event_team_link', 'team_triCode': 'event_team_tri_code', 
                        'result_secondaryType': 'event_detail', 'result_strength_code': 'goal_strength_code', 'result_strength_name': 'goal_strength_name',
                        'result_gameWinningGoal': 'goal_gwg', 'result_emptyNet': 'goal_empty_net', 'result_penaltySeverity': 'penalty_severity', 
                        'result_penaltyMinutes': 'penalty_minutes', 'gamePk': 'game_id'}
        
        event_df = event_df.rename(columns = column_names)
        
        ## Fixing event players' names        
        event_players = ['event_player_' + str(x + 1) for x in list(range(0, 4))]
        
        event_players = [x for x in event_players if x in event_df.columns]
        
        for player in event_players:
            
            event_df[player] = event_df[player].str.normalize('NFKD').str.encode('ascii', errors='ignore').str.decode('utf-8').str.upper()
            
            replace_dict = {'ALEXANDRE ': 'ALEX ', 'ALEXANDER ': 'ALEX ', 'CHRISTOPHER ': 'CHRIS '}
        
            event_df[player] = event_df[player].replace(to_replace = replace_dict)
            
            event_df[player] = event_df[player].map(correct_names_dict).fillna(event_df[player])

            name_split = event_df[player].str.split(' ')

            event_df[player + '_api_name'] = name_split.str[0] + '.' + name_split.str[1]

            event_df[player + '_api_name'] = event_df[player + '_api_name'].map(api_names_dict).fillna(event_df[player + '_api_name'])
            
        ## Changing event type names
        
        types_dict = {'BLOCKED_SHOT': 'BLOCK', 'BLOCKEDSHOT': 'BLOCK', 'MISSED_SHOT': 'MISS', 'FACEOFF': 'FAC',
                      'PENALTY': 'PENL', 'GIVEAWAY': 'GIVE', 'TAKEAWAY': 'TAKE', 'MISSEDSHOT': 'MISS'}
        
        event_df.event_type = event_df.event_type.map(types_dict).fillna(event_df.event_type)
        
        ## Adding additional time columns
        event_df.period = event_df.period.astype(int)
        
        event_df.period_time = event_df.period_time.astype(str)
        
        time_split = event_df.period_time.astype(str).str.split(':')
        
        event_df['period_seconds'] = time_split.str[0].astype(int) * 60 + time_split.str[1].astype(int)
        
        event_df['game_seconds'] = np.where(event_df.period_type != 'SHOOTOUT', (event_df.period - 1) * 1200, 3900)
        
        ## Creating opposition team columns, after swapping out the teams in blocked shots
        
        columns = ['team_id', 'team_name', 'team_link', 'team_tri_code']
        
        for column in columns:
            
            teams_dict = dict()
            
            teams_list = event_df['event_' + column].unique()
            
            teams_dict.update({teams_list[1]: teams_list[2],
                               teams_list[2]: teams_list[1]})
            
            event_df['event_' + column] = np.where(event_df.event_type == 'BLOCK',
                                                   event_df['event_' + column].map(teams_dict).fillna(event_df['event_' + column]),
                                                   event_df['event_' + column])
            
            event_df['opp_' + column] = event_df['event_' + column].map(teams_dict)
        
        ## Swapping out event players 1 and 2 for blocked shots
        players_dict = dict()
        
        players_dict.update({'event_player_1': 'event_player_2',
                             'event_player_2': 'event_player_1'})
        
        columns = ['link', 'type', 'id']
        
        for column in columns:
            
            players_dict.update({'event_player_1_' + column: 'event_player_2_' + column,
                                 'event_player_2_' + column: 'event_player_1_' + column})
            
        mask = event_df.event_type == 'BLOCK'
        
        event_df.update(event_df.loc[mask].rename(players_dict, axis = 1))
        
        ## Adding play ID column
        
        event_df['play_id'] = event_df.game_id.astype(int).astype(str) + '000'
        event_df['play_id'] = event_df.play_id.astype(int) + event_df.event_idx.astype(int)

        event_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)

        ## Add the game-level dataframe to the list of dataframes to be concatenated
        concat_list.append(event_df)
    
        if _print == True:
            
            print_game_id_time(game_start, game_id)
        
    if concat_list == []:

        events_df = pd.DataFrame()

    else:
        
        concat_start = time.perf_counter()
        
        if _print == True:
            
            print('\nStarting concatenation')
        
        ## Concatenate the game-level dataframes into one big dataframe
        events_df = pd.concat(concat_list, ignore_index = True)
    
        ## Rearranging columns
        columns = ['season', 'game_id', 'play_id', 'event_idx', 'event_idx_nhl', 'event_id_nhl', 'event_type', 'period', 'period_type', 'period_seconds',
                   'game_seconds', 'event_type_name', 'event_description', 'event_team_id', 'event_team_name', 'event_team_link',
                   'event_team_tri_code', 'event_player_1_type', 'event_player_1_id', 'event_player_1','event_player_1_link', 'event_player_1_api_name',
                   'event_player_2_type', 'event_player_2_id', 'event_player_2', 'event_player_2_link', 'event_player_2_api_name', 'event_player_3_type',
                   'event_player_3_id', 'event_player_3', 'event_player_3_link', 'event_player_3_api_name', 'event_player_4_type', 'event_player_4_id',
                   'event_player_4', 'event_player_4_link', 'event_player_4_api_name', 'opp_team_id', 'opp_team_name', 'opp_team_link', 'opp_team_tri_code',
                   'coords_x', 'coords_y', 'goals_home', 'goals_away', 'period_num', 'period_time', 'period_time_remaining', 'datetime',
                   'event_detail', 'goal_strength_code', 'goal_strength_name', 'goal_gwg', 'goal_empty_net', 'penalty_severity', 'penalty_minutes']
        
        columns = [x for x in columns if x in events_df.columns]

        events_df = events_df[columns]
        
        if _print == True:
            
            print_concat_finish(concat_start)

    if debug == True:

          if bad_game_list != []:

                print(f"{bad_game_list} did not have API info")

    if _print == True:

        print_number_of_games(game_ids, scrape_start)
    
    else:
        
        return events_df
        
## Scrape API game info
def scrape_api_game_info(game_ids, full_scrape = False, debug = False, _print = False):
    """
    Function to scrape game info data from the NHL API. Takes a game ID or a list of game IDs. Does not work for years before 2010.
    
    Can be used on its own, or as part of the larger full scrape function
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    full_scrape -- If False, scrapes the API for each game ID. Give the API response as a JSON otherwise (default False)
    debug -- If True, prints the game ID before attempting to scrape the data (default False)
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)

    if _print == True:
    
        print_scrape_intro(game_ids, 'API game info')
    
    ## Important lists
    bad_game_list = list()
    concat_list = list()
    
    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
            
        if full_scrape == False:
        
            s = s_session()

            response = s.get(str('https://statsapi.web.nhl.com/api/v1/game/' + str(game_id) + '/feed/live'), timeout = .5).json()
            
        else:
            
            response = full_scrape
        
        if response['gameData'] == []:
            
            if debug == True:
            
                print(f"{game_id} doesn't have API game info")
                
            bad_game_list.append(game_id)
            
            continue
        
        info_list = list()
        
        for key in response['gameData'].keys():
            
            info = pd.json_normalize(response['gameData'][key], sep = '_')

            if key == 'game':

                column_names = {'pk': 'game_id', 'type': 'game_type'}

                info = info.rename(columns = column_names)

                info_list.append(info)

            if key == 'datetime':

                column_names = {'dateTime': 'game_start_time', 'endDateTime': 'game_end_time'}

                info = info.rename(columns = column_names)

                info_list.append(info)

            if key == 'venue':

                column_names = {'id': 'game_venue_id', 'name': 'game_venue_name', 'link': 'game_venue_link'}

                info = info.rename(columns = column_names)

                info_list.append(info)
        
        info_df = pd.concat(info_list, axis = 1)
        
        ## Creating initial teams dataframe
        teams_dict = dict()
                
        for key in response['gameData']['teams'].keys():

            team_df = pd.json_normalize(response['gameData']['teams'][key], sep = '_')

            column_names = {'id': 'team_id', 'name': 'team_name', 'link': 'team_link', 'abbreviation': 'team_abbr', 'triCode': 'team_tri_code',
                            'teamName': 'team_mascot', 'locationName': 'team_location', 'firstYearOfPlay': 'team_first_year_of_play',
                            'shortName': 'team_short_name', 'officialSiteUrl': 'team_site_url', 'franchiseId': 'team_franchise_id',
                            'active': 'team_franchise_active', 'venue_id': 'team_venue_id', 'venue_name': 'team_venue_name',
                            'venue_link': 'team_venue_link', 'venue_city': 'team_venue_city', 'venue_timeZone_id': 'team_venue_tz_id',
                            'venue_timeZone_offset': 'team_venue_tz_offset', 'venue_timeZone_tz': 'team_venue_tz_name',
                            'division_id': 'team_division_id', 'division_name': 'team_division_name', 'division_link': 'team_division_link',
                            'conference_id': 'team_conference_id', 'conference_name': 'team_conference_name', 'conference_link': 'team_conference_link', 
                            'franchise_link': 'team_franchise_link', 'franchise_teamName': 'team_franchise_mascot', 'division_shortName': 'team_div_short_name'}

            team_df = team_df.rename(columns = column_names).drop('franchise_franchiseId', axis = 1)

            team_df['status'] = key

            teams_dict.update({key: team_df})

        teams_list = list()

        merge_dict = {'home': 'away', 'away': 'home'}

        for key, value in merge_dict.items():

            column_names = {x: 'opp_' + x for x in teams_dict[value].columns}
            
            teams_list.append(teams_dict[key].merge(teams_dict[value].rename(columns = column_names),
                                                    left_index = True, right_index = True))
            
        teams_df = pd.concat(teams_list)
        teams_df = pd.concat([teams_df, info_df], axis = 1).reset_index(drop = True)

        teams_df['game_id'] = game_id
        
        teams_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)

        concat_list.append(teams_df)
        
        if _print == True:
            
            print_game_id_time(game_start, game_id)
    
    if concat_list != []:
    
        game_info = pd.concat(concat_list, ignore_index = True)
    
    if debug == True:
        
        if bad_game_list != []:

            print(f"{bad_game_list} did not have API game info")
            
    if _print == True:
        
        print_number_of_games(game_ids, scrape_start)
    
    if concat_list == []:
        
        return pd.DataFrame()
    
    else:
        
        return game_info

## Scrape API roster info
def scrape_api_rosters(game_ids, full_scrape = False, debug = False, _print = False):
    """
    Function to scrape roster data from the NHL API. Takes a game ID or a list of game IDs. Does not work for years before 2010.
    
    Can be used on its own, or as part of the larger full scrape function
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    full_scrape -- If False, scrapes the API for each game ID. Give the API response as a JSON otherwise (default False)
    debug -- If True, prints the game ID before attempting to scrape the data (default False)
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)

    if _print == True:
    
        print_scrape_intro(game_ids, 'API roster info')
    
    ## Important lists
    bad_game_list = list()
    concat_list = list()
    
    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
            
        if full_scrape == False:
            
            s = s_session()

            response = s.get(str('https://statsapi.web.nhl.com/api/v1/game/' + str(game_id) + '/feed/live'), timeout = .5).json()
            
        else:
            
            response = full_scrape
        
        if response['gameData']['players'] == []:
            
            if debug == True:
            
                print(f"{game_id} doesn't have API roster data")

        players_list = list()
        
        for player in response['gameData']['players'].keys():

            df = pd.json_normalize(response['gameData']['players'][player], meta = 'gameData', sep = '_')

            column_names = {'id': 'player_id', 'fullName': 'player_name', 'link': 'player_link', 'firstName': 'player_first_name',
                            'lastName': 'player_last_name', 'primaryNumber': 'player_number', 'birthDate': 'player_dob',
                            'currentAge': 'player_current_age', 'birthCity': 'player_birth_city', 'birthStateProvince': 'player_birth_state_province', 
                            'birthCountry': 'player_birth_country', 'nationality': 'player_nationality', 'height': 'player_height', 'weight': 'player_weight', 
                            'active': 'player_active', 'alternateCaptain': 'player_alternate_captain', 'captain': 'player_captain', 'rookie': 'player_rookie', 
                            'shootsCatches': 'player_shoots_catches', 'rosterStatus': 'player_roster_status', 'currentTeam_id': 'player_current_team_id',
                            'currentTeam_name': 'player_current_team_name', 'currentTeam_link': 'player_current_team_link',
                            'currentTeam_triCode': 'player_current_team_tri_code', 'primaryPosition_code': 'player_position_code',
                            'primaryPosition_name': 'player_position_name', 'primaryPosition_type': 'player_position_type',
                            'primaryPosition_abbreviation': 'player_position_abbr'}

            df = df.rename(columns = column_names)

            df['game_id'] = game_id

            players_list.append(df)
            
        players_df = pd.concat(players_list)
        
        ## Fixing players' names        
        players_df.player_name = players_df.player_name.str.normalize('NFKD').str.encode('ascii', errors='ignore').str.decode('utf-8')#.str.upper()

        replace_dict = {'ALEXANDRE ': 'ALEX ', 'ALEXANDER ': 'ALEX ', 'CHRISTOPHER ': 'CHRIS '}
        
        players_df.player_name = players_df.player_name.replace(to_replace = replace_dict)
        
        players_df.player_name = players_df.player_name.map(correct_names_dict).fillna(players_df.player_name)
        
        name_split = players_df.player_name.str.split(' ')
        
        players_df.player_first_name = name_split.str[0]
        
        players_df.player_last_name = name_split.str[1]

        players_df['api_name'] = name_split.str[0] + '.' + name_split.str[1]

        players_df.api_name = players_df.api_name.map(api_names_dict).fillna(players_df.api_name)

        players_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)
        
        concat_list.append(players_df)
            
        if _print == True:

            print_game_id_time(game_start, game_id)
            
    if concat_list != []:
        
        concat_start = time.perf_counter()
        
        if _print == True:
            
            print('\nStarting concatenation')
        
        roster_df = pd.concat(concat_list, ignore_index = True)
        
        columns = ['season', 'game_id', 'player_id', 'player_name', 'api_name', 'player_current_team_name', 'player_current_team_tri_code', 'player_number',
                   'player_roster_status', 'player_position_code', 'player_position_type', 'player_position_abbr', 'player_position_name', 
                   'player_first_name', 'player_last_name', 'player_dob', 'player_height', 'player_weight', 'player_shoots_catches', 
                   'player_nationality', 'player_birth_city', 'player_birth_state_province', 'player_birth_country', 'player_rookie',
                   'player_current_age', 'player_captain', 'player_alternate_captain', 'player_link', 'player_current_team_id',
                   'player_current_team_link', 'player_active']
        
        columns = [x for x in columns if x in roster_df.columns]

        roster_df = roster_df[columns]
    
    if debug == True:
        
        if bad_game_list != []:

            print(f"{bad_game_list} did not have API roster info")
            
    if _print == True:
        
        print_concat_finish(concat_start)
        
        print_number_of_games(game_ids, scrape_start)
        
    if concat_list == []:
        
        return pd.DataFrame()
    
    else:
        
        return roster_df

## Scrape API boxscores
def scrape_api_boxscores(game_ids, debug = False, _print = False):
    """
    Function to boxscore stats and roster data from the NHL API. Does not work for years before 2010.
    
    Returns a dataframe of boscore stats and roster information for the game ID(s) specified.
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)

    if _print == True:
    
        print_scrape_intro(game_ids, 'API boxscore')
    
    ## Important lists
    concat_list = list()
    
    s = s_session()

    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
            
        response = s.get(f'https://statsapi.web.nhl.com/api/v1/game/{game_id}/boxscore', timeout = .5).json()
        
        game_concat_list = list()

        teams_list = ['home', 'away']

        for team in teams_list:

            team_concat_list = list()

            for player in response['teams'][team]['players'].values():

                team_concat_list.append(pd.json_normalize(player, sep = '_'))

            team_df = pd.concat(team_concat_list, ignore_index = True)

            team_df['game_id'] = game_id

            for key, value in response['teams'][team]['team'].items():

                team_df['team_' + key] = value

            game_concat_list.append(team_df)

        game_df = pd.concat(game_concat_list, ignore_index = True)

        game_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)
        
        ## Clean dataframe at the game level
        
        new_cols = list()
        
        for column in game_df.columns:
            
            column = column.replace('stats_', '')
            
            column = column.replace('person_', '')
            
            new_cols.append(column)
        
        column_names = dict(zip(game_df.columns, new_cols))
        
        game_df = game_df.rename(columns = column_names)
        
        column_names = {'season': 'season', 'game_id': 'game_id', 'team_name': 'team_name', 'id': 'player_id', 'fullName': 'player_name', 'jerseyNumber': 'player_jersey',
                        'position_abbreviation': 'player_position_abbr', 'skaterStats_timeOnIce': 'skater_toi', 'skaterStats_goals': 'skater_goals',
                        'skaterStats_assists': 'skater_assists', 'skaterStats_shots': 'skater_shots', 'skaterStats_hits': 'skater_hits',
                        'skaterStats_powerPlayGoals': 'skater_power_play_goals', 'skaterStats_powerPlayAssists': 'skater_power_play_assists',
                        'skaterStats_penaltyMinutes': 'skater_penalty_minutes', 'skaterStats_faceOffPct': 'skater_faceoff_win_percent',
                        'skaterStats_faceoffTaken': 'skater_faceoffs_taken', 'skaterStats_takeaways': 'skater_takaways',
                        'skaterStats_giveaways': 'skater_giveaways', 'skaterStats_shortHandedGoals': 'skater_shortHanded_goals',
                        'skaterStats_shortHandedAssists': 'skater_shorthanded_assists', 'skaterStats_blocked': 'skater_shots_blocked',
                        'skaterStats_plusMinus': 'skater_plus_minus', 'skaterStats_evenTimeOnIce': 'skater_even_toi',
                        'skaterStats_powerPlayTimeOnIce': 'skater_power_play_toi', 'skaterStats_shortHandedTimeOnIce': 'skater_shorthanded_toi',
                        'goalieStats_timeOnIce': 'goalie_toi', 'goalieStats_assists': 'goalie_goals', 'goalieStats_assists': 'goalie_assists',
                        'goalieStats_pim': 'goalie_penalty_minutes', 'goalieStats_shots': 'goalie_shots', 'goalieStats_saves': 'goalie_saves',
                        'goalieStats_powerPlaySaves': 'goalie_power_play_saves', 'goalieStats_shortHandedSaves': 'goalie_shorthanded_saves',
                        'goalieStats_evenSaves': 'goalie_even_saves', 'goalieStats_shortHandedShotsAgainst': 'goalie_shorthanded_shots_against',
                        'goalieStats_evenShotsAgainst': 'goalie_even_shots_against', 'goalieStats_powerPlayShotsAgainst': 'goalie_power_play_shots_against',
                        'goalieStats_decision': 'goalie_decision', 'goalieStats_savePercentage': 'goalie_save_percentage',
                        'goalieStats_powerPlaySavePercentage': 'goalie_power_play_save_percentage',
                        'goalieStats_shortHandedSavePercentage': 'goalie_shorthanded_save_percentage',
                        'goalieStats_evenStrengthSavePercentage': 'goalie_even_save_percentage', 'firstName': 'player_first_name',
                        'lastName': 'player_last_name', 'link': 'player_link', 'birthDate': 'player_dob', 'birthCity': 'player_birth_city', 
                        'birthStateProvince': 'player_birth_state_province', 'birthCountry': 'player_birth_country', 'nationality': 'player_nationality', 
                        'currentAge': 'player_current_age', 'height': 'player_height', 'weight': 'player_weight', 'active': 'player_active',
                        'roster_status': 'player_roster_status', 'shootsCatches': 'player_shoots_catches', 'team_id': 'team_id', 'link': 'player_link', 
                        'primaryNumber': 'player_primary_number', 'captain': 'player_captain', 'alternate_captain': 'player_alternate_captain',
                        'rookie': 'player_rookie', 'position_code': 'player_position_code', 'position_name': 'player_position_name',
                        'position_type': 'player_position_type', 'currentTeam_id': 'current_team_id', 'currentTeam_name': 'current_team_name',
                        'currentTeam_link': 'current_team_link', 'primaryPosition_code': 'player_primary_position_code',
                        'primaryPosition_name': 'player_primary_position_name', 'primaryPosition_type': 'player_primary_position_type',
                        'primaryPosition_abbreviation': 'player_primary_position_abbr'}
        
        game_df = game_df.rename(columns = column_names)
        
        columns = [x for x in column_names.values() if x in game_df.columns]
        
        game_df = game_df[columns]
        
        concat_list.append(game_df)
        
        if _print == True:

            print_game_id_time(game_start, game_id)
    
    if concat_list == []:
        
        boxscore_df = pd.DataFrame()
    
    else:
        
        concat_start = time.perf_counter()
        
        if _print == True:

            print('\nStarting concatenation')
        
        boxscore_df = pd.concat(concat_list, ignore_index = True)
        
        if _print == True:
            
            print_concat_finish(concat_start)

            print_number_of_games(game_ids, scrape_start)
            
    return boxscore_df        

## Scrape HTML shifts
def scrape_html_shifts(game_ids, _print = False, debug = False, full_scrape = False):
    """
    Function to scrape shifts data from the html endpoint. Returns a dictionary of shifts and changes dataframes
    
    Returns dictionary with the following attributes:
    'shifts': dataframe of shifts information
    'changes': dataframe of changes, to be used for merging with the events data
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)
    
    if _print == True:
    
        print_scrape_intro(game_ids, 'HTML shifts')
    
    ## Important lists
    shifts_concat = list()
    changes_concat = list()
    
    scrape_dict = dict()
    
    s = s_session()

    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
            
        html_season_id, html_game_id = convert_ids(game_id)
        
        urls_dict = {'home': f'http://www.nhl.com/scores/htmlreports/{html_season_id}/TH0{html_game_id}.HTM',
                     'away': f'http://www.nhl.com/scores/htmlreports/{html_season_id}/TV0{html_game_id}.HTM'}
        
        concat_list = list()

        for team, url in urls_dict.items():

            response = requests.get(url)

            soup = BeautifulSoup(response.content.decode('ISO-8859-1'), 'lxml', multi_valued_attributes = None)

            team_name = soup.find('td', {'align':'center', 'class':'teamHeading + border'}).get_text()

            players = soup.find_all('td', {'class':['playerHeading + border', 'lborder + bborder']})
            
            players_dict = dict()

            for player in players:

                data = player.get_text()

                if ', ' in data:

                    name = data.split(',')

                    number = name[0].split(' ')[0].strip()

                    last_name = name[0].split(' ')[1].strip()

                    first_name = name[1].strip()

                    full_name = f'{first_name} {last_name}'

                    players_dict[full_name] = dict()

                    players_dict[full_name]['number'] = number

                    players_dict[full_name]['name'] = full_name

                    players_dict[full_name]['shifts'] = list()

                else:

                    players_dict[full_name]['shifts'].extend([data])

            for key in players_dict.keys():

                length = int(len(np.array(players_dict[key]['shifts'])) / 5)

                column_names = {0: 'shift_number', 1: 'period', 2: 'shift_start', 3: 'shift_end', 4: 'duration'}

                player_df = pd.DataFrame(np.array(players_dict[key]['shifts']).reshape(length, 5)).rename(columns = column_names)

                player_df['player_name'] = players_dict[key]['name']

                player_df['player_jersey'] = players_dict[key]['number']

                player_df['team'] = team_name

                player_df['venue'] = team

                concat_list.append(player_df)

        shifts_df = pd.concat(concat_list, ignore_index = True)

        shifts_df['start_time'] = shifts_df.shift_start.str.split('/').str[0]

        shifts_df['end_time'] = shifts_df.shift_end.str.split('/').str[0]

        shifts_df['period'] = np.where(shifts_df.period == 'OT', 4, shifts_df.period).astype(int)

        start_time_split = shifts_df.start_time.astype(str).str.split(':')
        duration_time_split = shifts_df.duration.astype(str).str.split(':')

        condition = ~shifts_df.shift_end.str.contains('\xa0')
        value = shifts_df.end_time
        default_value = (pd.to_datetime((60 * start_time_split.str[0].astype(int)) +
                                        (start_time_split.str[1].astype(int)) + 
                                        (60 * duration_time_split.str[0].astype(int)) +
                                        (duration_time_split.str[1].astype(int)), unit = 's').dt.time).astype(str).str[4:]

        shifts_df['end_time'] = np.where(condition, value, default_value)

        replace_dict = {'ALEXANDRE ': 'ALEX ', 'ALEXANDER ': 'ALEX ', 'CHRISTOPHER ': 'CHRIS '}

        shifts_df.player_name = shifts_df.player_name.replace(to_replace = replace_dict)

        shifts_df.player_name = shifts_df.player_name.map(correct_names_dict).fillna(shifts_df.player_name)

        condition = pd.to_datetime(shifts_df.start_time).dt.time > pd.to_datetime(shifts_df.end_time).dt.time
        value = '20:00'
        default_value = shifts_df.end_time

        shifts_df.end_time = np.where(condition, value, default_value)

        shifts_df['goalie'] = np.where(shifts_df.player_name.isin(goalie_names), 1, 0)

        group_list = ['team', 'period']

        shifts_df['period_goalies'] = shifts_df.groupby(group_list)['goalie'].transform('sum')

        condition = (shifts_df.goalie == 1) & (shifts_df.start_time != '0:00') & (shifts_df.period_goalies == 1)
        value = '0:00'
        default_value = shifts_df.start_time

        shifts_df.start_time = np.where(condition, value, default_value)

        cond_1 = ((pd.to_datetime(shifts_df.start_time).dt.time < datetime(2021, 6, 10, 18, 0, 0).time()) &
                  (shifts_df.period != 3) & 
                  (shifts_df.period != 4) & 
                  (shifts_df.goalie == 1) & 
                  (shifts_df.period_goalies == 1))

        value_1 = '20:00'

        cond_2 = ((pd.to_datetime(shifts_df.start_time).dt.time < datetime(2021, 6, 10, 13, 0, 0).time()) &
                  (shifts_df.period != 4) & 
                  (shifts_df.goalie == 1) &
                  (shifts_df.period_goalies == 1))

        value_2 = '20:00'

        conditions = [cond_1, cond_2]
        values = [value_1, value_2]

        shifts_df.end_time = np.select(conditions, values, shifts_df.end_time)

        shifts_df.start_time = shifts_df.start_time.str.strip()
        shifts_df.end_time = shifts_df.end_time.str.strip()

        columns = ['start_time', 'end_time', 'duration']

        for column in columns: 

            time_split = shifts_df[column].astype(str).str.split(':')

            shifts_df[f'{column}_seconds'] = 60 * time_split.str[0].astype(int) + time_split.str[1].astype(int)

        group_dict = {'on': ['team', 'period', 'start_time', 'start_time_seconds'], 
                      'off': ['team', 'period', 'end_time', 'end_time_seconds']}

        times_dict = {'on': {'start_time': 'time'}, 
                      'off': {'end_time': 'time'}}

        changes_dict = dict()

        for change_type, group_list in group_dict.items():

            column_names = times_dict[change_type]

            sort_list = ['team', 'period', 'time']

            if change_type == 'on':

                df = shifts_df.groupby(group_list, as_index = False)\
                        .agg(players_on = ('player_name', ', '.join),
                             players_on_numbers = ('player_jersey', ', '.join),
                             number_on = ('player_name', 'count'))\
                        .rename(columns = column_names)\
                        .sort_values(by = sort_list)

            if change_type == 'off':

                df = shifts_df.groupby(group_list, as_index = False)\
                        .agg(players_off = ('player_name', ', '.join),
                             players_off_numbers = ('player_jersey', ', '.join),
                             number_off = ('player_name', 'count'))\
                        .rename(columns = column_names)\
                        .sort_values(by = sort_list)

            changes_dict.update({change_type: df.reset_index(drop = True)})

        merge_list = ['team', 'period', 'time']

        changes_on = changes_dict['on'].merge(changes_dict['off'], on = merge_list, how = 'left')
        changes_off = changes_dict['off'].merge(changes_dict['on'], on = merge_list, how = 'left')

        changes_df = pd.concat([changes_on, changes_off], ignore_index = True)\
                            .sort_values(by = ['period', 'start_time_seconds'], ascending = True)\
                            .drop_duplicates().reset_index(drop = True)
        
        changes_df.start_time_seconds = changes_df.start_time_seconds.fillna(1200)
        
        changes_df.end_time_seconds = changes_df.end_time_seconds.fillna(0)
        
        changes_df = changes_df.sort_values(by = ['period', 'start_time_seconds'], ascending = True).reset_index(drop = True)

        time_split = changes_df.time.astype(str).str.split(':')

        changes_df['period_seconds'] = 60 * time_split.str[0].astype(int) + time_split.str[1].astype(int)

        changes_df['game_seconds'] = (changes_df.period - 1) * 1200 + changes_df.period_seconds

        changes_df.team = np.where(changes_df.team=='CANADIENS MONTREAL', 'MONTREAL CANADIENS', changes_df.team)
        
        shifts_df['game_id'] = game_id
        shifts_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)
        
        changes_df['game_id'] = game_id
        changes_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)
        
        #shifts_df['shift_index'] = shifts_df.groupby(['period', 'start_time_seconds', 'team']).ngroup()
        
        #changes_df['shift_index'] = changes_df.groupby(['period', 'start_time_seconds', 'team']).ngroup() + 1
        
        shifts_concat.append(shifts_df)
        
        changes_concat.append(changes_df)
        
        if _print == True:

            print_game_id_time(game_start, game_id)
    
    concat_start = time.perf_counter()
    
    if _print == True:
        
        print('\nStarting concatenation')
        
    if shifts_concat == []:
        
        shifts_df = pd.DataFrame()
        
    else:
        
        shifts_df = pd.concat(shifts_concat, ignore_index = True)
        
        columns = ['game_id', 'team', 'venue', 'player_name', 'player_jersey', 'shift_number', 'period',
                   'shift_start', 'shift_end', 'duration', 'start_time', 'end_time', 'start_time_seconds',
                   'end_time_seconds', 'duration_seconds', 'goalie']
        
        columns = [x for x in columns if x in shifts_df.columns]
                   
        shifts_df = shifts_df[columns]
        
    if changes_concat == []:
        
        changes_df = pd.DataFrame()
        
    else:
        
        changes_df = pd.concat(changes_concat, ignore_index = True)
                   
        columns = ['game_id', 'team', 'period', 'time', 'start_time_seconds', 'players_on', 'players_on_numbers', 'number_on',
                   'end_time_seconds', 'players_off', 'players_off_numbers', 'number_off', 'period_seconds', 'game_seconds']
        
        columns = [x for x in columns if x in changes_df.columns]
        
        changes_df = changes_df[columns]
        
    if _print == True:
        
        print_concat_finish(concat_start)
        
    scrape_dict.update({'shifts': shifts_df, 
                       'changes': changes_df})
    
    if _print == True:
        
        print_number_of_games(game_ids, scrape_start)
    
    return scrape_dict

## Scrape API shifts
def scrape_api_shifts(game_ids, debug = False, _print = False):
    """
    Function to scrape shifts data from the API endpoint. Returns a dictionary of shifts and changes dataframes
    
    Returns dictionary with the following attributes:
    'shifts': dataframe of shifts information
    'changes': dataframe of changes, to be used for merging with the events data
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)
    
    if _print == True:
    
        print_scrape_intro(game_ids, 'API shifts')
    
    ## Important lists
    shifts_concat = list()
    changes_concat = list()
    
    scrape_dict = dict()
    
    s = s_session()

    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
    
        url = f'https://api.nhle.com/stats/rest/en/shiftcharts?cayenneExp=gameId={game_id}'

        response = s.get(url).json()['data']
        
        column_names = {'gameId': 'game_id', 'teamName': 'team_name', 'playerId': 'player_id', 'firstName': 'player_first_name',
                        'lastName': 'player_last_name', 'shiftNumber': 'player_shift_number', 'period': 'period', 'startTime': 'shift_start_time',
                        'endTime': 'shift_end_time','duration': 'shift_duration', 'eventNumber': 'event_number', 'teamAbbrev': 'team_name_abbr',
                        'teamId': 'team_id','hexValue': 'hex_value', 'typeCode': 'type_code', 'eventDescription': 'event_description',
                        'eventDetails': 'event_details', 'detailCode': 'detail_code'}
        
        shifts_df = pd.json_normalize(response, sep = '_').rename(columns = column_names)

        if shifts_df.empty:

            continue

        if 'player_first_name' not in shifts_df.columns:

            continue
        
        shifts_df['player_name'] = shifts_df.player_first_name + ' ' + shifts_df.player_last_name
        
        shifts_df['season'] = str(game_id)[:4] + str(int(str(game_id)[:4]) + 1)
        
        times_list = ['start_time', 'end_time', 'duration']
        
        columns = ['shift_' + x for x in times_list]

        for column in columns:

            shifts_df[column] = np.where(shifts_df[column] == '', np.nan, shifts_df[column])

            shifts_df[column].fillna('0:0', inplace = True)

            if column == 'shift_duration':

                shifts_df[column] = shifts_df[column].fillna(np.nan)

                time_split = shifts_df[column].astype(str).str.split(':')

                shifts_df[column + '_seconds'] = 60 * time_split.str[0].astype(float) + time_split.str[1].astype(float)

            else:

                time_split = shifts_df[column].astype(str).str.split(':')

                shifts_df[column + '_period_seconds'] = 60 * time_split.str[0].astype(int).astype(int) + time_split.str[1].astype(int)

                shifts_df[column + '_game_seconds'] = (shifts_df.period.astype(int) - 1) * 1200 + shifts_df[column + '_period_seconds'].astype(int)
                
        shifts_df['is_event'] = np.where(pd.isna(shifts_df.shift_duration), 1, 0)
        
        on_group = ['season', 'game_id', 'team_name', 'team_id', 'team_name_abbr', 'period', 'shift_start_time',
                    'shift_start_time_period_seconds', 'shift_start_time_game_seconds']

        off_group = ['season', 'game_id', 'team_name', 'team_id', 'team_name_abbr', 'period', 'shift_end_time',
                     'shift_end_time_period_seconds', 'shift_end_time_game_seconds']

        group_dicts = {'on': {'game_seconds': 'shift_start_time_game_seconds',
                              'time': 'shift_start_time',
                              'group_list': on_group},
                       'off' : {'game_seconds': 'shift_end_time_game_seconds',
                                'time': 'shift_end_time',
                                'group_list': off_group}}

        changes_dict = dict()

        for change_type, group_dict in group_dicts.items():

            sort_list = ['team_name_abbr', group_dict['game_seconds']]

            df = shifts_df[shifts_df.is_event != 1]

            df = df.astype({'player_id': 'str'})

            if change_type == 'on':

                df = df.groupby(group_dict['group_list'], as_index = False)\
                        .agg(players_on = ('player_name', tuple),
                             players_on_ids = ('player_id', tuple),
                             number_on = ('player_name', 'count'))\
                        .sort_values(by = sort_list)

            if change_type == 'off':

                df = df.groupby(group_dict['group_list'], as_index = False)\
                        .agg(players_off = ('player_name', tuple),
                             players_off_ids = ('player_id', tuple),
                             number_off = ('player_name', 'count'))\
                        .sort_values(by = sort_list)

            df['time'] = df[group_dict['time']]

            df['game_seconds'] = df[group_dict['game_seconds']]

            changes_dict.update({change_type: df.reset_index(drop = True)})

            merge_list = ['season', 'game_id', 'team_name', 'team_id', 'team_name_abbr', 'period', 'time', 'game_seconds']

        changes_on = changes_dict['on'].merge(changes_dict['off'], on = merge_list, how = 'left')
        changes_off = changes_dict['off'].merge(changes_dict['on'], on = merge_list, how = 'left')

        changes_df = pd.concat([changes_on, changes_off], ignore_index = True)\
                            .sort_values(by = ['game_seconds'], ascending = True)\
                            .drop_duplicates().reset_index(drop = True)
        
        #columns = [x for x in column_names.values() if x in shifts_df.columns]
        
        #shifts_df = shifts_df[columns]
        
        shifts_concat.append(shifts_df)
        
        changes_concat.append(changes_df)
        
        if _print == True:

            print_game_id_time(game_start, game_id)
    
    concat_start = time.perf_counter()
    
    if _print == True:
        
        print('\nStarting concatenation')
        
    if shifts_concat == []:
        
        shifts_df = pd.DataFrame()
        
    else:
        
        shifts_df = pd.concat(shifts_concat, ignore_index = True)
        
    if changes_concat == []:
        
        changes_df = pd.DataFrame()
        
    else:
        
        changes_df = pd.concat(changes_concat, ignore_index = True)
        
    if _print == True:
        
        print_concat_finish(concat_start)
        
    scrape_dict.update({'shifts': shifts_df,
                        'changes': changes_df})
    if _print == True:
        
        print_number_of_games(game_ids, scrape_start)
    
    return scrape_dict

## Scrape HTML events
def scrape_html_events(game_ids, debug = False, _print = False, roster_df = pd.DataFrame()):
    """
    Function to scrape events data from the html endpoint. Returns a dataframe of events data
    
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    debug -- 
    _print -- If True, prints information during scrape (default False)
    rosters -- 
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)
    
    if _print == True:
    
        print_scrape_intro(game_ids, 'HTML events')
    
    ## Important lists
    concat_list = list()
    
    s = s_session()

    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
            
        html_season_id, html_game_id = convert_ids(game_id)

        url = f'http://www.nhl.com/scores/htmlreports/{html_season_id}/PL0{html_game_id}.HTM'

        response = s.get(url, timeout = .5)

        soup = BeautifulSoup(response.content.decode('ISO-8859-1'), 'lxml')

        tds = soup.find_all("td", {"class": re.compile('.*bborder.*')})

        stripped_html = hs_strip_html(tds)

        length = int(len(stripped_html) / 8)

        column_names = {0: 'event_idx', 1: 'period', 2: 'strength', 3: 'time', 4: 'event', 5: 'description', 6: 'away_skaters', 7: 'home_skaters'}

        events_df = pd.DataFrame(np.array(stripped_html).reshape(length, 8)).rename(columns = column_names)

        potential_names = soup.find_all('td', {'align':'center', 'style':'font-size: 10px;font-weight:bold'})

        teams_dict = dict()

        for name in potential_names:

            data = name.get_text()

            if ('Away Game') in data or ('tr./Away') in data:

                away = re.split('Match|Game', data)[0]

                teams_dict.update({'away': away})

                break

        for name in potential_names:

            data = name.get_text()

            if ('Home Game') in data or ('Dom./Home') in data:

                home = re.split('Match|Game', data)[0]

                teams_dict.update({'home': home})

                break

        for team_type, team_name in teams_dict.items():

            events_df[team_type + '_skaters'] = events_df[team_type + '_skaters'].str.replace('\n', '')

            events_df[team_type + '_team'] = team_name

            events_df[team_type + '_team_abbreviated'] = events_df[team_type + '_skaters'][0].split(' ')[0]

        time_split = events_df.time.astype(str).str.split(':')

        events_df['original_time'] = events_df.time

        events_df['time'] = time_split.str[0] + ':' + time_split.str[1].str[:2]

        events_df = events_df[events_df.period != 'Per'].reset_index(drop = True)

        events_df['event_team'] = events_df.description.str.split(' ').str[0].str.split('\xa0').str[0]

        home_team_name = events_df.home_team_abbreviated.iloc[0]
        away_team_name = events_df.away_team_abbreviated.iloc[0]
        teams_list = [home_team_name, away_team_name]

        events_df.event_team = np.where(np.isin(events_df.event_team, teams_list), events_df.event_team, '\xa0')

        cond_1 = events_df.event_team == ' '
        value_1 = '\xa0'

        cond_2 = events_df.event_team == home_team_name
        value_2 = away_team_name

        cond_3 = events_df.event_team == away_team_name
        value_3 = home_team_name

        conditions = [cond_1, cond_2, cond_3]
        values = [value_1, value_2, value_3]

        events_df['opp_team'] = np.select(conditions, values, '\xa0')

        events_df['event_player_str'] = events_df.description.str.findall('(#)(\d\d)|(#)(\d)|(-) (\d\d)|(-) (\d)').astype(str).str\
                                                    .replace('#', '', regex = False).str.replace(',', '', regex = False).str\
                                                    .replace('-', '', regex = False).str.replace("'", '', regex = False).str\
                                                    .replace('(', '', regex = False).str.replace(')', '', regex = False).str\
                                                    .replace('[', '', regex = False).str.replace(']', '', regex = False).str\
                                                    .replace(' +', ' ', regex = True).str.strip()
                                                    #.apply(lambda x: re.sub(' +', ' ', x)).str.strip()

        player_split = events_df.event_player_str.str.split(' ')

        players = ['event_player_' + str(x) for x in list(range(1, 4))]

        for idx, player in enumerate(players):

            events_df[player] = player_split.str[idx]

        events_df.event_player_2 = np.where(events_df.description.str.contains('Drawn By'), 
                                            events_df.description.astype(str).str\
                                                .split('Drawn By').astype(str).str[1].str\
                                                .split('#').astype(str).str[1].str\
                                                .split(' ').astype(str).str[0].str.strip(), 
                                            events_df.event_player_2)

        events_df.event_player_3 = np.where(events_df.description.str.contains('Served By'), '\xa0', events_df.event_player_3)

        conditions = [(~pd.isna(events_df.event_player_1) & (events_df.event_player_1 != '') & (events_df.event == 'FAC')), 
                      (~pd.isna(events_df.event_player_1) & (events_df.event_player_1 != '') & (events_df.event != 'FAC'))]

        values = [events_df.away_team_abbreviated + events_df.event_player_1.astype(str), 
                  events_df.event_team + events_df.event_player_1.astype(str)]

        events_df.event_player_1 = np.select(conditions, values, events_df.event_player_1)

        event_list = ['BLOCK', 'HIT', 'PENL']

        conditions = [(~pd.isna(events_df.event_player_2) & (events_df.event_player_2 != '') & (events_df.event == 'FAC')),
                      (~pd.isna(events_df.event_player_2) & (events_df.event_player_2 != '') & (events_df.event != 'FAC') &
                       (events_df.event.isin(event_list))),
                      (~pd.isna(events_df.event_player_2) & (events_df.event_player_2 != '') & (events_df.event != 'FAC') &
                       (~events_df.event.isin(event_list)))]

        values = [events_df.home_team_abbreviated + events_df.event_player_2.astype(str),
                  events_df.opp_team + events_df.event_player_2.astype(str),
                  events_df.event_team + events_df.event_player_2.astype(str)]

        events_df.event_player_2 = np.select(conditions, values, events_df.event_player_2)

        events_df.event_player_3 = np.where((~pd.isna(events_df.event_player_3) & (events_df.event_player_3 != '')), 
                                            events_df.event_team + events_df.event_player_3.astype(str), 
                                            events_df.event_player_3)
        
        if roster_df.empty:
            
            roster_df = scrape_html_rosters(game_id)
            
        else:
            
            roster_df = roster_df
            
        roster_df['team_abbreviated'] = np.where(roster_df.team == 'home', home_team_name, away_team_name)
        
        roster_df['teamnum'] = roster_df.team_abbreviated + roster_df.player_jersey.astype(str)
        
        name_dict = dict(zip(roster_df.teamnum, roster_df.player_full_name))
        
        api_name_dict = dict(zip(roster_df.player_full_name, roster_df.api_name))
        
        players = ['event_player_' + str(x) for x in list(range(1, 4))]
        
        for player in players:
            
            events_df[player] = events_df[player].map(name_dict).fillna(events_df[player])
            
            events_df[player + '_api_name'] = events_df[player].map(api_name_dict).fillna(events_df[player])
        
        game_date = soup.find_all('td', {'align':'center', 'style':'font-size: 10px;font-weight:bold'})[2].get_text()
        
        events_df['game_date'] = pd.to_datetime(game_date)
        
        events_df.period = np.where(events_df.period == '', 1, events_df.period).astype(int)
        
        events_df.time = np.where(np.logical_or(events_df.time == '', pd.isna(events_df.time)), '0:00', events_df.time)
        
        time_split = events_df.time.astype(str).str.replace('-', '').astype(str).str.split(':')
        
        events_df['period_seconds'] = 60 * time_split.str[0].astype(int) + time_split.str[1].astype(int)
        
        events_df['game_seconds'] = 1200 * (events_df.period - 1) + events_df.period_seconds
        
        events_df['season'] = html_season_id
        
        events_df['game_id'] = game_id
        
        events_df = events_df.replace('CANADIENS MONTREAL', 'MONTREAL CANADIENS', regex = False)
        
        if game_id == 2007020003:
            
            events_df.game_date = events_df.game_date + pd.Timedelta(days = 1)
        
        events_df.event_player_1 = np.where(np.logical_or(events_df.description.str.contains('team', case = False), 
                                                          events_df.description.str.contains('bench', case = False)),
                                            'BENCH', events_df.event_player_1)
        
        type_list = ['home', 'away']
        
        for team_type in type_list:
            
            events_df[team_type + '_skater_count'] = events_df[team_type + '_skaters'].str.count('[A-Z]')
            
        home_penalty = events_df.home_skater_count > events_df.home_skater_count.shift(-1)
        away_penalty = events_df.away_skater_count > events_df.away_skater_count.shift(-1)
        
        cond_1 = ((events_df.event == 'PENL') & (events_df.event_team == '') &
                  (events_df.description.str.contains('bench', case = False) &
                   home_penalty))
        
        value_1 = events_df.home_team_abbreviated
        
        cond_2 = ((events_df.event == 'PENL') & (events_df.event_team == '') &
                  (events_df.description.str.contains('bench', case = False) &
                   away_penalty))
        
        value_2 = events_df.away_team_abbreviated
        
        conditions = [cond_1, cond_2]
        values = [value_1, value_2]
        
        events_df.event_team = np.select(conditions, values, events_df.event_team)
        
        concat_list.append(events_df)
        
        if _print == True:

            print_game_id_time(game_start, game_id)
    
    concat_start = time.perf_counter()
    
    if _print == True:
        
        print('\nStarting concatenation')
    
    if concat_list == []:
        
        events_df = pd.DataFrame()
    
    events_df = pd.concat(concat_list, ignore_index = True)
    
    columns = ['season', 'game_id', 'event_idx', 'period', 'period_seconds', 'game_seconds', 'event',
               'description', 'event_team', 'event_player_1', 'event_player_2', 'event_player_3', 'opp_team', 
               'home_team', 'home_team_abbreviated', 'home_skaters', 'away_team', 'away_team_abbreviated', 'away_skaters',
               'strength', 'home_skater_count', 'away_skater_count', 'time', 'original_time', 'game_date', 'event_player_1_api_name', 
               'event_player_2_api_name', 'event_player_3_api_name', 'event_player_str']
    
    columns = [x for x in columns if x in events_df.columns]
    
    events_df = events_df[columns]
    
    if _print == True:
        
        print_concat_finish(concat_start)
        
        print_number_of_games(game_ids, scrape_start)
        
    return events_df

## Full scrape function
def full_scrape(game_ids, scrape_type = ['all'], debug = False, _print = True):
    """
    Function to scrape some or all of the NHL API game data. Takes a game ID or a list of game IDs. Does not work for years before 2010.
    
    Returns a dictionary of dataframes, with the following attributes:
        'events' : Dataframe of play-by-play data
        'api_rosters' : Dataframe of rosters from the play-by-play API endpoint
        'game_info' : Dataframe of information about the teams playing in the game
        'html_rosters' : Dataframe of information from the html endpoint
        'api_boxscores' : Dataframe of information from the API boxscore
        'html_shifts' : Dataframe of information from the html shifts endpoint
        'html_changes' : Dataframe of information from the html shifts endpoint
        'html_events' : Dataframe of information from the html events endpoint
    Keyword arguments:
    game_ids -- Single game ID or list of game IDs
    scrape_type -- Pass a list of a combination of the dictionary keys (default 'all')
    _print -- If True, prints information during scrape (default False)
    """
    
    ## Starting timer for scrape start
    scrape_start = time.perf_counter()

    good_list = ['all', 'events', 'api_rosters', 'game_info', 'html_rosters', 'api_boxscores', 'html_shifts', 'html_changes', 'html_events',
                    'api_events', 'api_shifts', 'api_changes']
    
    scrape_type = convert_to_list(scrape_type)

    for x in scrape_type:

        if x not in good_list:

            raise Exception(f"Sorry '{x}' is not a supported scrape type")

    final_dict = dict()
    
    ## Convert game IDs to list if given a single game ID
    game_ids = convert_to_list(game_ids)

    if _print == True:
    
        print_scrape_intro(game_ids, 'full')
    
    ## Important lists
    bad_game_dict = dict()
    events_concat = list()
    api_rosters_concat = list()
    game_info_concat = list()
    html_rosters_concat = list()
    api_boxscores_concat = list()
    html_shifts_concat = list()
    html_changes_concat = list()
    api_events_concat = list()
    html_events_concat = list()
    api_shifts_concat = list()
    api_changes_concat = list()
    
    s = s_session()

    for game_id in game_ids:
        
        game_start = time.perf_counter()
        
        if debug == True:
            
            print(game_id)
        
        response = s.get(str('https://statsapi.web.nhl.com/api/v1/game/' + str(game_id) + '/feed/live'), timeout = .5).json()       
        
        if scrape_type == ['all']:
        
            api_events_df = scrape_api_events(game_id, full_scrape = response)
            if ~api_events_df.empty:
                api_events_concat.append(api_events_df)

            api_rosters_df = scrape_api_rosters(game_id, full_scrape = response)
            if ~api_rosters_df.empty:
                api_rosters_concat.append(api_rosters_df)    

            game_info_df = scrape_api_game_info(game_id, full_scrape = response)
            if ~game_info_df.empty:
                game_info_concat.append(game_info_df)
                
            html_rosters_df = scrape_html_rosters(game_id)
            if ~html_rosters_df.empty:
                html_rosters_concat.append(html_rosters_df)

            api_boxscores_df = scrape_api_boxscores(game_id)
            if ~api_boxscores_df.empty:
                api_boxscores_concat.append(api_boxscores_df)

            html_shifts_dict = scrape_html_shifts(game_id)

            html_shifts_df = html_shifts_dict['shifts']
            html_changes_df = html_shifts_dict['changes']

            if ~html_shifts_df.empty:
                html_shifts_concat.append(html_shifts_df)

            if ~html_changes_df.empty:
                html_changes_concat.append(html_changes_df)

            html_events_df = scrape_html_events(game_id, roster_df = html_rosters_df)
            if ~html_events_df.empty:
                html_events_concat.append(html_events_df)

            api_shifts_dict = scrape_api_shifts(game_id)

            api_shifts_df = api_shifts_dict['shifts']
            api_changes_df = api_shifts_dict['changes']

            if ~api_shifts_df.empty:
                api_shifts_concat.append(api_shifts_df)

            if ~api_changes_df.empty:
                api_changes_concat.append(api_changes_df)

        if 'api_events' in scrape_type:
            
            api_events_df = scrape_api_events(game_id, full_scrape = response)
            if ~api_events_df.empty:
                api_events_concat.append(api_events_df)
                
        if 'game_info' in scrape_type:
            
            game_info_df = scrape_api_game_info(game_id, full_scrape = response)
            if ~game_info_df.empty:
                game_info_concat.append(game_info_df)
                
        if 'api_rosters' in scrape_type:
            
            api_rosters_df = scrape_api_rosters(game_id, full_scrape = response)
            if ~api_rosters_df.empty:
                api_rosters_concat.append(api_rosters_df)
                
        if 'html_rosters' in scrape_type:
            
            html_rosters_df = scrape_html_rosters(game_id)
            if ~html_rosters_df.empty:
                html_rosters_concat.append(html_rosters_df)

        if 'api_boxscores' in scrape_type:

            api_boxscores_df = scrape_api_boxscores(game_id)
            if ~api_boxscores_df.empty:
                api_boxscores_concat.append(api_boxscores_df)

        if 'html_shifts' in scrape_type:

            html_shifts_dict = scrape_html_shifts(game_id)

            html_shifts_df = html_shifts_dict['shifts']
            html_changes_df = html_shifts_dict['changes']

            if ~html_shifts_df.empty:
                html_shifts_concat.append(html_shifts_df)

            if ~html_changes_df.empty:
                html_changes_concat.append(html_changes_df)

        if 'html_events' in scrape_type:

            if 'html_rosters' in scrape_type:

                html_events_df = scrape_html_events(game_id, roster_df = html_rosters_df)

            else:

                html_events_df = scrape_html_events(game_id)

            if ~html_events_df.empty:
                html_events_concat.append(html_events_df)

        if 'api_shifts' in scrape_type:
            api_shifts_dict = scrape_api_shifts(game_id)

            api_shifts_df = api_shifts_dict['shifts']
            api_changes_df = api_shifts_dict['changes']

            if ~api_shifts_df.empty:
                api_shifts_concat.append(api_shifts_df)

            if ~api_changes_df.empty:
                api_changes_concat.append(api_changes_df)

        if _print == True:

            print_game_id_time(game_start, game_id)
    
    if _print == True:

        print('\nStarting concatenation')

    concat_start = time.perf_counter()

    if api_events_concat != []:

        api_events_df = pd.concat(api_events_concat, ignore_index = True)
        
        final_dict.update({'api_events': api_events_df})
        
    if api_rosters_concat != []:
        
        api_rosters_df = pd.concat(api_rosters_concat, ignore_index = True)
        
        final_dict.update({'api_rosters': api_rosters_df})
        
    if game_info_concat != []:
        
        game_info_df = pd.concat(game_info_concat, ignore_index = True)
        
        final_dict.update({'game_info': game_info_df})
        
    if html_rosters_concat != []:
        
        html_rosters_df = pd.concat(html_rosters_concat, ignore_index = True)
        
        final_dict.update({'html_rosters': html_rosters_df})

    if api_boxscores_concat != []:

        api_boxscores_df = pd.concat(api_boxscores_concat, ignore_index = True)

        final_dict.update({'api_boxscores': api_boxscores_df})

    if html_shifts_concat != []:

        html_shifts_df = pd.concat(html_shifts_concat, ignore_index = True)

        final_dict.update({'html_shifts': html_shifts_df})

    if html_changes_concat != []:

        html_changes_df = pd.concat(html_changes_concat, ignore_index = True)

        final_dict.update({'html_changes': html_changes_df})

    if html_events_concat != []:

        html_events_df = pd.concat(html_events_concat, ignore_index = True)

        final_dict.update({'html_events': html_events_df})

    if api_shifts_concat != []:

        api_shifts_df = pd.concat(api_shifts_concat, ignore_index = True)

        final_dict.update({'api_shifts': api_shifts_df})

    if api_changes_concat != []:

        api_changes_df = pd.concat(api_changes_concat, ignore_index = True)

        final_dict.update({'api_changes': api_changes_df})
        
    if _print == True:

        print_concat_finish(concat_start)

        print_number_of_games(game_ids, scrape_start)
    
    return final_dict

## Combine and scrape both roster types (will likely be deprecated)
# def scrape_rosters(game_ids, _batch = False, outer_print = True, _inner_print = False):
    
#     scrape_start = time.perf_counter()

#     game_ids = convert_to_list(game_ids)
    
#     print_scrape_intro(game_ids, 'full rosters')
    
#     if len(game_ids) > 100:
        
#         _batch = True
    
#     if _batch == True:
        
#         if outer_print == True:
            
#             print(f'Batch downloading and cleaning {len(game_ids)} games')
        
#         html_start = time.perf_counter()

#         html = scrape_html_rosters(game_ids, _print = _inner_print)

#         if outer_print == True:

#             message = f'Finished scraping {len(game_ids)} games from the HTML source in'

#             end_timer(time.perf_counter() - html_start, message)

#         api_start = time.perf_counter()

#         api = scrape_api_rosters(game_ids, _print = _inner_print)
        
#         if outer_print == True:

#             message = f'Finished scraping {len(game_ids)} games from the API in'

#             end_timer(time.perf_counter() - api_start, message)

#         if outer_print == True:
            
#             intro_time = time.perf_counter() - scrape_start
            
#             end_timer(intro_time, message = f'Finished with downloading and initial preparation of data sets in')
        
#         merge_start = time.perf_counter()
        
#         roster_df = html.merge(api, on = ['game_id', 'api_name'], how = 'outer')
        
#         if outer_print == True:
            
#             merge_time = time.perf_counter() - merge_start
            
#             end_timer(merge_time, message = f'Finished with merge of {len(game_ids)} games in')
            
#             scrape_time =  time.perf_counter() - scrape_start
            
#             end_timer(scrape_time, message = f'Finished with batch download in')
        
#         return roster_df
        
#     else:
        
#         rosters_list = []
        
#         for game_id in game_ids:

#             game_start = time.perf_counter()

#             html = scrape_html_rosters(game_id, _print = _inner_print)
#             api = scrape_api_rosters(game_id, _print = _inner_print)

#             game_df = api.merge(html, on = 'api_name', how = 'outer')

#             rosters_list.append(game_df)

#             if outer_print == True:

#                 print_game_id_time(game_start, game_id)
        
#         roster_df = pd.concat(rosters_list, ignore_index = True)
     
#         if outer_print == True:

#             print_number_of_games(game_ids, scrape_start)
    
#         return roster_df

## Scraping api_rosters
# def scrape_api_rosters(game_ids, _print = True, output_type = 'dataframe'):
#     """
#     Game IDs can be a list or for a single game. Returns a dataframe of roster information from the API. By default, prints status update to console.
#     This will be nested within a larger roster scraping function but can also be accessed on its own.
#     """
    
#     scrape_start = time.perf_counter()
    
#     game_ids = convert_to_list(game_ids)
    
#     if _print == True:
        
#         print_scrape_intro(game_ids, 'API rosters')
    
#     s = s_session()
    
#     api_list = []
    
#     for game_id in game_ids:
        
#         players_list = []
        
#         game_start = time.perf_counter()
        
#         url_dict = {'api_events' : f'https://statsapi.web.nhl.com/api/v1/game/{game_id}/feed/live',
#                     'api_boxscore' : f'https://statsapi.web.nhl.com/api/v1/game/{game_id}/boxscore'}
        
#         ## Making API calls
#         boxscore = s.get(url_dict['api_boxscore'], timeout = 2.5).json()
#         events = s.get(url_dict['api_events'], timeout = 2.5).json()
        
#         ## Creating data frame with game information, including time and date
#         new_cols = {'pk' : 'game_id',
#                     'type' : 'game_type',
#                     'dateTime' : 'game_date',
#                     'abstractGameState' : 'game_status',
#                     'detailedState' : 'detailed_game_status',
#                     'startTimeTBD' : 'start_time_TBD'}
        
#         game_info = pd.json_normalize(events['gameData']['game']).rename(columns = new_cols)
        
#         time_info = pd.json_normalize(events['gameData']['datetime']).rename(columns = new_cols)
        
#         time_info['start_time'] = pd.to_datetime(time_info.game_date).dt.time
        
#         time_info['game_date'] = pd.to_datetime(time_info.game_date, infer_datetime_format = True).dt.date
        
#         game_state = pd.json_normalize(events['gameData']['status']).rename(columns = new_cols).drop(['codedGameState', 'statusCode'], axis = 1)
        
#         ## Creating data frame for venue info for the game
#         try:
            
#             venue_info = pd.json_normalize(events['gameData']['teams']['home']['venue'], sep = '_').add_prefix('game_venue_')
            
#         except KeyError:
            
#             try:
                
#                 venue_info = pd.json_normalize(events['gameData']['venue'], sep = '_').add_prefix('game_venue_')
                
#             except KeyError:
                
#                 venue_info = pd.DataFrame()
        
#         game_info = pd.concat([game_info, time_info, game_state, venue_info], axis = 1)
        
#         empty_list = []
        
#         ## Looping through home and away teams 
#         for team in events['gameData']['teams'].keys():
            
#             team_info = events['gameData']['teams'][team].copy()
            
#             if 'venue' in team_info.keys():
                
#                 del team_info['venue']
            
#             ## Getting team information to add to player data
#             team_info = pd.json_normalize(team_info, sep = '_')
            
#             ## Creating dataframe of team venue information
#             try:
#                 team_venue = pd.json_normalize(events['gameData']['teams'][team]['venue'], sep = '_').add_prefix('team_venue_')
                
#             except:
#                 team_venue = pd.DataFrame()
            
#             ## Creating meta data dataframe by combiniing team information, venue information, and team venue information
#             meta_data = pd.concat([game_info, team_info, team_venue], axis = 1)
            
#             ## Creating new column names for meta data data frame. These will be renamed later
#             new_cols = ['id', 'name', 'link', 'abbreviation', 'locationName', 'firstYearOfPlay', 'shortName', 'officialSiteUrl']

#             new_cols_dict = {}

#             for col in new_cols:

#                 new_cols_dict.update({col : 'team_' + str(col)})

#             new_cols_dict.update({'active' : 'franchise_active'})

#             meta_data.rename(columns = new_cols_dict, inplace = True) 
            
#             ## Getting player IDs to loop through box score information
#             player_ids = boxscore['teams'][team]['players'].keys()
            
#             if (len(player_ids) == 0) & (len(events['gameData']['players']) == 0):
                
#                 players_list.append(meta_data)
                
#                 empty_list.append(1)
                
#             else:
                
#                 ## Looping through players
#                 for player_id in player_ids:

#                     ## Creating a dataframe for each player
#                     player_info = (boxscore['teams'][team]['players'][player_id]['person'])

#                     player_df = pd.json_normalize(player_info, sep = '_',)

#                     ## Addiing meta data to player dataframe
#                     player_df = pd.concat([player_df, meta_data], axis = 1)

#                     ## Adding player dataframe to box score dataframe
#                     players_list.append(player_df.copy(deep = True))
                    
#         if len(empty_list) == 2:

#             print(f'Sorry, {game_id} does not have any player information')
        
#         else:
            
#             api_game_df = pd.concat(players_list, ignore_index = True)
            
#             ##Dictionary of column names to change
#             new_cols = {'jerseyNumber' : 'player_jersey',
#                         'person_id' : 'player_id',
#                         'id' : 'player_id',
#                         'person_fullName' : 'player_full_name',
#                         'fullName' : 'player_full_name',
#                         'firstName' : 'player_first_name', 
#                         'lastName' : 'player_last_name', 
#                         'link' : 'player_link',
#                         'person_link' : 'player_link',
#                         'person_firstName' : 'player_first_name',
#                         'person_lastName' : 'player_last_name',
#                         'person_primaryNumber' : 'player_primary_jersey',
#                         'primaryNumber' : 'player_primary_jersey',
#                         'person_birthDate' : 'player_birth_date',
#                         'birthDate' : 'player_birth_date',
#                         'person_birthCity' : 'player_birth_city',
#                         'birthCity' : 'player_birth_city',
#                         'person_birthStateProvince' : 'player_birth_state',
#                         'birthStateProvince' : 'player_birth_state',
#                         'person_birthCountry' : 'player_country',
#                         'birthCountry' : 'player_country',
#                         'person_nationality' : 'player_nationality',
#                         'nationality' : 'player_nationality',
#                         'person_height' : 'player_height',
#                         'height' : 'player_height', 
#                         'weight' : 'player_weight', 
#                         'active' : 'player_active', 
#                         'shootsCatches' : 'player_shoots', 
#                         'rosterStatus' : 'player_roster_status',
#                         'person_weight' : 'player_weight',
#                         'person_active' : 'player_active',
#                         'person_rookie' : 'player_rookie',
#                         'person_shootsCatches' : 'player_shoots',
#                         'person_rosterStatus' : 'player_roster_status',
#                         'primaryPosition_abbreviation': 'player_primary_position',
#                         'primaryPosition_type' : 'player_primary_position_type',
#                         'person_primaryPosition_abbreviation' : 'player_primary_position',
#                         'person_primaryPosition_type' : 'player_primary_position_type',
#                         'position_abbreviation' : 'player_position',
#                         'position_type' : 'player_position_type',
#                         'triCode' : 'team_tri_code',
#                         'team_name' : 'team_full_name',
#                         'teamName' : 'team_mascot',
#                         'team_locationName' : 'team_city',
#                         'team_firstYearOfPlay' : 'team_start_season',
#                         'team_shortName' : 'team_short_name',
#                         'team_officialSiteUrl' : 'team_url',
#                         'franchiseId' : 'franchise_id',
#                         'franchise_teamName' : 'franchise_team_name',
#                         'game_venue_timeZone_id' : 'game_venue_tz_id',
#                         'game_venue_timeZone_offset' : 'game_venue_tz_offset',
#                         'game_venue_timeZone_tz' : 'game_venue_tz',
#                         'team_venue_timeZone_id' : 'team_venue_tz_id',
#                         'team_venue_timeZone_offset' : 'team_venue_tz_offset',
#                         'team_venue_timeZone_tz' : 'team_venue_tz',
#                         'stats_goalieStats_timeOnIce' : 'goalie_toi',
#                         'stats_goalieStats_assists' : 'goalie_assists',
#                         'stats_goalieStats_goals' : 'goalie_goals',
#                         'stats_goalieStats_pim' : 'goalie_pim',
#                         'stats_goalieStats_shots' : 'goalie_shots_against',
#                         'stats_goalieStats_saves' : 'goalie_saves',
#                         'stats_goalieStats_powerPlaySaves' : 'goalie_pp_saves',
#                         'stats_goalieStats_shortHandedSaves' : 'goalie_sh_saves',
#                         'stats_goalieStats_evenSaves' : 'goalie_ev_saves',
#                         'stats_goalieStats_shortHandedShotsAgainst' : 'goalie_sh_shots_against',
#                         'stats_goalieStats_evenShotsAgainst' : 'goalie_ev_shots_against',
#                         'stats_goalieStats_powerPlayShotsAgainst' : 'goalie_pp_shots_against',
#                         'stats_goalieStats_decision' : 'goalie_decision',
#                         'stats_goalieStats_savePercentage' : 'goalie_save_percentage',
#                         'stats_goalieStats_powerPlaySavePercentage' : 'goalie_pp_save_percentage',
#                         'stats_goalieStats_shortHandedSavePercentage' : 'goalie_sh_save_percentage',
#                         'stats_goalieStats_evenStrengthSavePercentage' : 'goalie_ev_save_percentage',
#                         'currentAge' : 'player_current_age', 
#                         'alternateCaptain' : 'player_captain_alternate',
#                         'currentTeam_id' : 'player_current_team_id',
#                         'currentTeam_name' : 'player_current_team_name', 
#                         'currentTeam_link' : 'player_current_team_link',
#                         'person_currentAge' : 'player_current_age',
#                         'person_alternateCaptain' : 'player_captain_alternate',
#                         'person_captain' : 'player_captain',
#                         'person_currentTeam_id' : 'player_current_team_id',
#                         'person_currentTeam_name' : 'player_current_team_name',
#                         'person_currentTeam_link' : 'player_current_team_link'}

#             api_game_df.rename(columns = new_cols, inplace = True)

#             if 'player_full_name' in list(api_game_df.columns):

#                 # Max Pacioretty doesn't exist in ESPN in 2009-2010, sadly.
#                 shorter_names_dict = {'AlEXANDRE ' : 'ALEX ',
#                                       'ALEXANDER ' : 'ALEX ',
#                                       'CHRISTOPHER ' : 'CHRIS '}

#                 for longer_name, shorter_name in shorter_names_dict.items():

#                     api_game_df.player_full_name = np.where(np.isin(longer_name, api_game_df.player_full_name.str.upper()),
#                                                             shorter_name,
#                                                             api_game_df.player_full_name.str.upper())

#                 # List of names and fixed from Evolving Hockey Scraper.
#                 api_game_df.player_full_name = api_game_df.player_full_name.map(correct_names_dict).fillna(api_game_df.player_full_name)

#                 player_names = ['player_first_name', 'player_last_name']

#                 for idx, player_name in enumerate(player_names):

#                     api_game_df[player_name] = api_game_df.player_full_name.str.split(' ', 1).str[idx]

#                 api_game_df['api_name'] = api_game_df.player_first_name.str.upper() + '.' + api_game_df.player_last_name.str.upper()

#                 ## Correcting for double names
#                 double_names = {8480222 : 'SEBASTIAN.AHO2', 
#                                 8471221 : 'ALEX.PICARD2', 
#                                 8474744 : 'SEAN.COLLINS2', 
#                                 8478400 : 'COLIN.WHITE2', 
#                                 8476979 : 'ERIK.GUSTAFSSON2',
#                                 8482247 : 'MIKKO.LEHTONEN2'}


#                 api_game_df['api_name'] = api_game_df.player_id.map(double_names).fillna(api_game_df.api_name)

#             ## Dropping extra columns
#             drop_list = ['person_primaryPosition_code',
#                          'person_primaryPosition_name',
#                          'position_code',
#                          'position_name',
#                          'franchise_franchiseId']

#             api_game_df.drop(drop_list, axis = 1, inplace = True, errors = 'ignore')
            
#             api_list.append(api_game_df)
        
#         if _print == True:
            
#             print_game_id_time(game_start, game_id)
    
#     concat_start = time.perf_counter()

#     if _print == True:

#         print(f'Starting concatenation')

#     api_df = pd.concat(api_list, ignore_index = True)

#     if _print == True:

#         print_concat_finish(concat_start)

#         print_number_of_games(game_ids, scrape_start)
    
#     return api_df
